/* Task lists item */
.task-lists-item:drop(active) {
    background-color: @blue_2;
}

.task-blue {
    background: #264770;
}

.task-green {
    background: #345039;
}

.task-yellow {
    background: #5e5900;
}

.task-orange {
    background: #65461f;
}

.task-red {
    background: #5e0800;
}

.task-purple {
    background: #5c395c;
}

.task-brown {
    background: #512e12;
}

link {
    color: white;
}